/* 

Copyright (c) 1999-2001  Ulric Eriksson <ulric@siag.nu>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <stdio.h>
#include <stdlib.h>
#include <magick/api.h>
#include <magick/magick.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <limits.h>

#ifndef USHRT_MAX
#define USHRT_MAX 65535
#endif

int XMagickOldAlloc = 0;

static int rgb_shift = 11;

/* known colours are cached in a sorted list to avoid
   making multiple calls to the X server */
/* Beware though: alloc_color sorts on rgb, query_color does not sort */
typedef struct xmagick_color {
	XColor color;
	unsigned long rgb;	/* (up to) 24-bit colour */
} xmagick_color;

static xmagick_color *known_colors = NULL;
static XColor *known_pixels = NULL;

static int ncolors, maxcolors, lastcolor, lastpixel, npixels, maxpixels;

static void die(char *p)
{
	fprintf(stderr, "%s\n", p);
	exit(EXIT_FAILURE);
}

/* forget all cached colours */
static void free_colors(void)
{
	free(known_colors);
	known_colors = NULL;
	ncolors = 0;
	lastcolor = -1;
}

/* forget all cached pixels */
static void free_pixels(void)
{
	free(known_pixels);
	known_pixels = NULL;
	npixels = 0;
	lastpixel = -1;
}

/* call XAllocColor through the cache */
/* Does not fail. Returns closest match at worst. */
/* Algorithm:
   1. If same rgb value as last time, return same pixel value.
   2. If exact match found, return it.
   3. Call XAllocColor. If successful, return it.
   4. Return "anything".

   If N is the number of picture elements (width*height), M is the number
   of picture elements that are the same as the one before, C is the number
   of unique colours, we get O((N-M)*log(C) + C*C + M*1). Those come from
   number of table lookups + number of colour insertions + "same" colours.
*/
static void alloc_color1(Display *display, Colormap colormap, XColor *color)
{
	int i, j, lower, upper;
	int cr, cg, cb, d;
	unsigned long rgb;

	cr = color->red >> rgb_shift;
	cg = color->green >> rgb_shift;
	cb = color->blue >> rgb_shift;
	rgb = (cr << 16) + (cg << 8) + cb;

	/* take advantage of long stretches of identical pixels */
	if (lastcolor != -1 && known_colors[lastcolor].rgb == rgb) {
		color->pixel = known_colors[lastcolor].color.pixel;
		return;
	}
	/* binary search through known colours */
	lower = 0;
	upper = ncolors-1;
	while (lower <= upper) {
		i = (lower+upper)/2;
		d = rgb-known_colors[i].rgb;
		if (d == 0) {
			lastcolor = i;
			color->pixel = known_colors[i].color.pixel;
			return;
		} else if (d > 0) {
			lower = i+1;
		} else {
			upper = i-1;
		}
	}
	/* if we are here, then lower == upper+1 and no match found */
	if (ncolors >= maxcolors) {	/* too many colours */
		/* no reason to change lastcolor */
		color->pixel = 0;
		return;
	}
	if (XAllocColor(display, colormap, color) == 0) {
		int dr, dg, db, bd = INT_MAX, best = 0;
		/* linear search, yuck */
		for (j = 0; j < ncolors; j++) {
			XColor c = known_colors[j].color;
			dr = cr - (c.red >> rgb_shift);
			dg = cg - (c.green >> rgb_shift);
			db = cb - (c.blue >> rgb_shift);
			d = dr*dr+dg*dg+db*db;
			if (d < bd) {
				bd = d;
				best = j;
			}
		}
		color->pixel = known_colors[best].color.pixel;
	}
	/* insert new colour */
	if (known_colors[upper].rgb > rgb) i = upper;
	else i = lower;
	if (i < 0) i = 0;		/* before everything */
	if (i > ncolors) i = ncolors;	/* after everything */
	for (j = ncolors; j > i; j--) {
		known_colors[j] = known_colors[j-1];
	}
	known_colors[i].color = *color;
	known_colors[i].rgb = rgb;
	known_colors[i].color.pixel = color->pixel;
	lastcolor = i;
	ncolors++;
}

/* Algorithm:
   1. Combine r g b into an index in an rgb cube.
   2. If 0, return black.
   3. If cached, return pixel value.
   4. Call XAllocColor. If successful, return it.
   5. Go looking for close colours.

   Testing on a normal jpeg of a photo, this was 30% faster than alloc_color1.
*/
static unsigned long pixelcache[32768];
static unsigned long black;
static int pixelcache_ready = 0;
static void alloc_color2(Display *, Colormap, XColor *);

static void init_pixelcache(Display *d, Colormap cm)
{
	int i;
	int r, g, b;
	XColor color;
	Screen *s = DefaultScreenOfDisplay(d);

	black = BlackPixelOfScreen(s);
	for (i = 0; i < 32768; i++) {
		pixelcache[i] = black;
	}
	pixelcache_ready = 1;
	for (r = 0; r <= 31; r += 31) {
		for (g = 0; g <= 31; g += 31) {
			for (b = 0; b <= 31; b += 31) {
				color.red = r;
				color.green = g;
				color.blue = b;
				alloc_color2(d, cm, &color);
			}
		}
	}
}

static void alloc_color2(Display *display, Colormap colormap, XColor *color)
{
	int i, j, lower, upper;
	int cr, cg, cb, d;
	unsigned long rgb;
	int lastrgb = -1;
	int r, g, b, r0, g0, b0, r1, g1, b1;
	int o;

	if (!pixelcache_ready) init_pixelcache(display, colormap);

#if 0
	cr = color->red >> 11;
	cg = color->green >> 11;
	cb = color->blue >> 11;
#else
	cr = color->red * 31 / 65535;
	cg = color->green * 31 / 65535;
	cb = color->blue * 31 / 65535;
#endif
	rgb = (cr << 10) | (cg << 5) | cb;

	/* is it black? */
	if (rgb == 0) {
		color->pixel = black;
		return;
	}

	/* check if we have the value in the table already */
	if (pixelcache[rgb] != black) {
		color->pixel = pixelcache[rgb];
		return;
	}

	/* try to allocate the approximated colour */
	color->red = cr * 65535 / 31;
	color->green = cg * 65535 / 31;
	color->blue = cb * 65535 / 31;
	if (XAllocColor(display, colormap, color)) {
		pixelcache[rgb] = color->pixel;
		return;
	}

	/* go look for a similar colour */
	for (o = 1; o < 4; o++) {
		r0 = cr-o;
		if (r0 < 0) r0 = 0;
		r1 = cr+o;
		if (r1 > 31) r1 = 31;
		for (r = r0; r <= r1; r++) {
			g0 = cg-o;
			g1 = cg+o;
			if (g0 < 0) g0 = 0;
			if (g1 > 31) g1 = 31;
			for (g = g0; g <= g1; g++) {
				b0 = cb-o;
				b1 = cb+o;
				if (b0 < 0) b0 = 0;
				if (b1 > 31) b1 = 31;
				for (b = b0; b <= b1; b++) {
					i = (r << 10) | (g << 5) | b;
					if (pixelcache[i] != black) {
						color->pixel = pixelcache[i];
						pixelcache[rgb] = pixelcache[i];
						return;
					}
				}
			}
		}
	}

	/* give up */
	color->pixel = black;
}


static void alloc_color(Display *display, Colormap colormap, XColor *color)
{
	if (XMagickOldAlloc) alloc_color1(display, colormap, color);
	else alloc_color2(display, colormap, color);
}

/* call XQueryColor through the cache */
/* Algorithm (adapted from alloc_color1):
   1. If same pixel value as last time, return same rgb values.
   2. If exact match found, return it.
   3. Call XQueryColor and insert the new XColor into the sorted table.
*/
static void query_color(Display *display, Colormap colormap, XColor *color)
{
	int i, j, lower, upper;

	/* take advantage of long stretches of identical pixels */
	if (lastpixel != -1 && known_pixels[lastpixel].pixel == color->pixel) {
		*color = known_pixels[lastpixel];
		return;
	}

	/* binary search through known colours */
	lower = 0;
	upper = npixels-1;
	while (lower <= upper) {
		i = (lower+upper)/2;
		if (known_pixels[i].pixel == color->pixel) {
			*color = known_pixels[i];
			lastpixel = i;
			return;
		} else if (known_pixels[i].pixel < color->pixel) {
			lower = i+1;
		} else {
			upper = i-1;
		}
	}

	/* if we are here, then lower == upper+1 and no match found */
	if (npixels >= maxpixels) {	/* too many pixels */
		/* no reason to change lastpixel */
		*color = known_pixels[0];
		return;
	}

	XQueryColor(display, colormap, color);
	/* insert new pixel */
	if (known_pixels[upper].pixel > color->pixel) i = upper;
	else i = lower;
	if (i < 0) i = 0;		/* before everything */
	if (i > npixels) i = npixels;	/* after everything */
	for (j = npixels; j > i; j--) {
		known_pixels[j] = known_pixels[j-1];
	}
	known_pixels[i] = *color;
	lastpixel = i;
	npixels++;
}

/* Convert Image to XImage */
XImage *XMagickImageToXImage(Display *display, Image *image)
{
	XImage *ximage = NULL;
	int x, y;
	int screen = DefaultScreen(display);
	Visual *visual = DefaultVisual(display, screen);
	int format = ZPixmap;
	int offset = 0;
	char *data = NULL;
	Colormap colormap = DefaultColormap(display, screen);
	int bitmap_pad = XBitmapPad(display);
	int bytes_per_line = 0;
	unsigned int width = image->columns;
	unsigned int height = image->rows;
	unsigned int depth = DefaultDepth(display, screen);
	XColor color;
	unsigned long pixel;
	int i;
#if MagickLibVersion < 0x0520
	float *red, *green, *blue, *opacity;
#else
	PixelPacket *pp;
#endif

#if MagickLibVersion >= 0x0550
	/* RDB: Added Wed Dec 31 12:43:11 PST 2003 */
	ExceptionInfo	exception;
	GetExceptionInfo( &exception );
	maxcolors = GetNumberColors(image, NULL, &exception);

	if (exception.severity != UndefinedException)
				CatchException(&exception);
#else
	maxcolors = GetNumberColors(image, NULL);
#endif

	free_colors();
	known_colors = malloc(maxcolors*sizeof *known_colors);
	if (known_colors == NULL) die("Can't allocate color table\n");

#if MagickLibVersion < 0x0520

	red = malloc((width*height)*(sizeof *red));
	green = malloc((width*height)*(sizeof *green));
	blue = malloc((width*height)*(sizeof *blue));
	opacity = malloc((width*height)*(sizeof *opacity));

	if (red && green && blue && opacity) {
		GetPixels(image, red, green, blue, opacity);
		ximage = XCreateImage(display, visual, depth,
				format, offset, data,
				width, height, bitmap_pad, bytes_per_line);
		if (ximage == NULL) die("Can't create XImage");
		ximage->data = malloc(ximage->bytes_per_line*ximage->height);
		if (ximage->data == NULL) die("Can't allocate XImage data");
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				i = x+y*width;
				color.red = USHRT_MAX*red[i];
				color.green = USHRT_MAX*green[i];
				color.blue = USHRT_MAX*blue[i];
				alloc_color(display, colormap, &color);
				pixel = color.pixel;
				XPutPixel(ximage, x, y, pixel);
			}
		}
	}

	free(red);
	free(green);
	free(blue);
	free(opacity);

#else	/* ImageMagick 5 */

	pp = GetImagePixels(image, 0, 0, width, height);
	ximage = XCreateImage(display, visual, depth,
			format, offset, data,
			width, height, bitmap_pad, bytes_per_line);
	if (ximage == NULL) die("Can't create XImage\n");
	ximage->data = malloc(ximage->bytes_per_line*ximage->height);
	if (ximage->data == NULL) die("Can't allocate XImage data\n");
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			i = x+y*width;
			color.red = USHRT_MAX*pp[i].red/MaxRGB;
			color.green = USHRT_MAX*pp[i].green/MaxRGB;
			color.blue = USHRT_MAX*pp[i].blue/MaxRGB;
			alloc_color(display, colormap, &color);
			pixel = color.pixel;
			XPutPixel(ximage, x, y, pixel);
		}
	}
#endif
	free_colors();
	return ximage;
}

/* Convert XImage to Image */
Image *XMagickXImageToImage(Display *display, XImage *ximage)
{
	Image *image = NULL;
	XColor color;
	unsigned long pixel;
	int i;
	float *red, *green, *blue, *opacity;
	int x, y;
	int screen = DefaultScreen(display);
	Colormap colormap = DefaultColormap(display, screen);
	unsigned int width = ximage->width;
	unsigned int height = ximage->height;
#if MagickLibVersion >= 0x0520
	PixelPacket *pp;
#endif

	free_pixels();
	maxpixels = 65555;	/* arbitrary */
	known_pixels = malloc(maxpixels*sizeof *known_pixels);
#if MagickLibVersion < 0x0520
	red = malloc((width*height)*(sizeof *red));
	green = malloc((width*height)*(sizeof *green));
	blue = malloc((width*height)*(sizeof *blue));
	opacity = malloc((width*height)*(sizeof *opacity));

	if (red && green && blue && opacity) {
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				i = x+y*width;
				pixel = XGetPixel(ximage, x, y);
				color.pixel = pixel;
				query_color(display, colormap, &color);
				red[i] = (float)color.red/USHRT_MAX;
				green[i] = (float)color.green/USHRT_MAX;
				blue[i] = (float)color.blue/USHRT_MAX;
				opacity[i] = 1.0;
			}
		}
		image = CreateImage(width, height, red, green, blue, opacity);
	}

	free(red);
	free(green);
	free(blue);
	free(opacity);
#else	/* ImageMagick 5 */
	image = AllocateImage(NULL);
	image->columns = width;
	image->rows = height;
	i = 0;
	for (y = 0; y < height; y++) {
		pp = SetImagePixels(image, 0, y, width, 1);
		for (x = 0; x < width; x++) {
			pixel = XGetPixel(ximage, x, y);
			color.pixel = pixel;
			query_color(display, colormap, &color);
			pp[x].red = MaxRGB*color.red/USHRT_MAX;
			pp[x].green = MaxRGB*color.green/USHRT_MAX;
			pp[x].blue = MaxRGB*color.blue/USHRT_MAX;
			pp[x].opacity = OpaqueOpacity;
		}
		SyncImagePixels(image);
	}
#endif
	free_pixels();
	return image;
}

/* Set the number of bits to use when calculating the rgb value for
   alloc_color. Valid values are 1 to 8. Returns the new value.
*/
int XMagickRGBBits(int bits)
{
	if (bits < 1) bits = 1;
	else if (bits > 8) bits = 8;
	rgb_shift = 16-bits;
	return bits;
}

